"use strict";

var _babelHelpers = require("babel-runtime/helpers")["default"];

var _core = require("babel-runtime/core-js")["default"];

var HTTP = require("http");

module.exports = (function () {
  function PortMap() {
    _babelHelpers.classCallCheck(this, PortMap);

    this._ports = {};
    this._http = HTTP.request;
    HTTP.request = this._request.bind(this);
  }

  PortMap.prototype.map = function map(hostname, port) {
    this._ports[hostname] = port;
  };

  PortMap.prototype.unmap = function unmap(hostname) {
    delete this._ports.hostname;
  };

  PortMap.prototype._request = function _request(options, callback) {
    var hostname = options.hostname || options.host && options.host.split(":")[0] || "localhost";
    var port = options.port || options.host && options.host.split(":")[1] || 80;
    if (port === 80) {
      var mapped = this._find(hostname);
      if (mapped) {
        options = _core.Object.assign({}, options, {
          hostname: hostname,
          port: mapped
        });
      }
    }
    return this._http(options, callback);
  };

  PortMap.prototype._find = function _find(domain) {
    var _this = this;

    // Turn domain into a list of matches, from most to least specific, e.g.
    // 'foo.example.com' turns into:
    //
    // [ 'foo.example.test',
    //   '*.foo.example.test',
    //   '*.example.test',
    //   '*.test' ]
    var parts = domain.split(".");
    var domains = [domain];
    while (parts.length) {
      domains.push("*." + parts.join("."));
      parts.shift();
    }

    return domains.map(function (pattern) {
      return _this._ports[pattern];
    }).filter(function (port) {
      return port;
    })[0];
  };

  return PortMap;
})();
//# sourceMappingURL=port_map.js.map